/*
 * Cell Broadband Engine Performance Monitor
 *
 * (C) Copyright IBM Corporation 2001,2006
 *
 * Author: David Erb (djerb@us.ibm.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/types.h>

#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/reg.h>
#include <asm/spu.h>

#include "cbe_regs.h"
#include "interrupt.h"
#include "perfmon.h"


/* Most PM registers are write-only, so shadow the values written */

static struct cbe_shadow_regs {
	struct pm		group_control;
	struct pm		debug_bus_control;
	struct pm_status	pm_status;
	struct pm_control	pm_control;
	struct pm		pm_interval;
	struct pm		pm_ctr[NR_PHYS_CTRS];
	struct pm		pm_start_stop;
	struct pm07_control	pm07_control[NR_PHYS_CTRS*2];
	int			counter_value_in_latch;
} cbe_shadow_regs[MAX_CBE];

static struct cbe_shadow_regs *shadow[NR_CPUS] = {
	&cbe_shadow_regs[0], &cbe_shadow_regs[0],
	&cbe_shadow_regs[1], &cbe_shadow_regs[1]
};

/* When writing to write-only mmio addresses, save a shadow copy */
#define write_wo_mmio(register,x)		\
{						\
	shadow[cpu]->register.val = x;		\
	out_be32 (&pmd_regs->register.val, x);	\
}

/* Set the size of a physical counter */
void set_cbe_ctr_size(u32 cpu, u32 phys_ctr, u32 ctr_size)
{
	if (phys_ctr >= NR_PHYS_CTRS)
		return;

	switch (ctr_size) {
		case 16:
			shadow[cpu]->pm_control.sixteen_bit_ctr |=
						CBE_16BIT_CTR(phys_ctr);
			break;

		case 32:
			shadow[cpu]->pm_control.sixteen_bit_ctr &=
						~CBE_16BIT_CTR(phys_ctr);
			break;
	}
}


/* Return the size of a physical counter */
u32 get_cbe_ctr_size(u32 cpu, u32 phys_ctr)
{
	if (phys_ctr < NR_PHYS_CTRS)
		return 0;

	if (shadow[cpu]->pm_control.sixteen_bit_ctr & CBE_16BIT_CTR(phys_ctr))
		return 16;
	else
		return 32;
}


/* Read a physical counter (either 1 32-bit or 2 16-bit) */
u32 cbe_read_phys_ctr(u32 cpu, u32 phys_ctr)
{
	struct cbe_pmd_regs __iomem *pmd_regs;
	u32 val;

	if (phys_ctr >= NR_PHYS_CTRS)
		return 0;

	/*
	 * Read the latch or the actual counter, whichever is newer
	 */
	if (shadow[cpu]->counter_value_in_latch & phys_ctr) {
		val = shadow[cpu]->pm_ctr[phys_ctr].val;
	} else {
		pmd_regs = cbe_get_cpu_pmd_regs(cpu);
		val = in_be32(&pmd_regs->pm_ctr[phys_ctr].val);
	}

	return val;
}


/* Read a counter (either 32-bit or 16-bit) */
u32 cbe_read_ctr(u32 cpu, u32 ctr)
{
	u32 val;
	u32 phys_ctr;

	phys_ctr = ctr & (NR_PHYS_CTRS - 1);

	val = cbe_read_phys_ctr(cpu, phys_ctr);

	if (get_cbe_ctr_size(cpu, phys_ctr) == 16)
		val = (ctr < NR_PHYS_CTRS) ? (val >> 16) : (val & 0xffff);

	return val;
}


/* Set the value of a physical counter (either 1 32-bit or 2 16-bit) */
void cbe_write_phys_ctr(u32 cpu, u32 phys_ctr, u32 val)
{
	struct cbe_pmd_regs __iomem *pmd_regs;

	if (phys_ctr >= NR_PHYS_CTRS)
		return;

	pmd_regs = cbe_get_cpu_pmd_regs(cpu);
	out_be32(&(pmd_regs->pm_ctr[phys_ctr].val), val);

	/*
	 * Writing to a counter only writes to a latch.  The new value is not
	 * propagated to the actual counter until the performance monitor is
	 * enabled.
	 */
	shadow[cpu]->counter_value_in_latch |= phys_ctr;
	write_wo_mmio (pm_ctr[phys_ctr], val);
}


/* Set the value of a counter (either 32-bit or 16-bit) */
void cbe_write_ctr(u32 cpu, u32 ctr, u32 val)
{
	u32 phys_ctr;
	u32 phys_val;

	phys_ctr = ctr & (NR_PHYS_CTRS - 1);

	if (get_cbe_ctr_size(cpu, phys_ctr) == 16) {
		phys_val = cbe_read_phys_ctr(cpu, phys_ctr);

		if (ctr < NR_PHYS_CTRS)
			val = (val << 16) | (phys_val & 0xffff);
		else
			val = (val & 0xffff) | (phys_val & 0xffff0000);
	}

	cbe_write_phys_ctr (cpu, phys_ctr, val);
}

u32 cbe_clear_pm_interrupts(u32 cpu)
{
	struct cbe_pmd_regs __iomem *pmd_regs;
	struct cbe_iic_regs __iomem *iic_regs;
	u32 mask;

	pmd_regs = cbe_get_cpu_pmd_regs(cpu);
	iic_regs = cbe_get_cpu_iic_regs(cpu);

	/* 1) Read pm_status to clear the interrupt -- order is important! */
	mask = in_be32 (&pmd_regs->pm_status.val);

	/* 2) Write 1 to PMI bit to clear the interrupt */
	out_be64 (&iic_regs->iic_is, CBE_IIC_IS_PMI);

	return mask;
}

void cbe_enable_pm_interrupts(u32 cpu, u32 thread, u32 mask)
{
	struct cbe_pmd_regs __iomem *pmd_regs;
	struct cbe_iic_regs __iomem *iic_regs;
	struct cbe_iic_ir iic_ir;
	u32 node = cpu >> 1;

	pmd_regs = cbe_get_cpu_pmd_regs(cpu);
	iic_regs = cbe_get_cpu_iic_regs(cpu);

	/* Set which node and thread will handle the next interrupt */
	iic_ir.val = 0;
	iic_ir.dst_node_id = node;
	iic_ir.dst_unit_id = ((thread == 0) ? CBE_IIC_IR_PT0
					    : CBE_IIC_IR_PT1);

	out_be64 (&iic_regs->iic_ir.val, iic_ir.val);

	/* Enable pm interrupts */
	if (mask)
		write_wo_mmio (pm_status, mask);
}

void cbe_disable_pm_interrupts(u32 cpu)
{
	struct cbe_pmd_regs __iomem *pmd_regs;

	pmd_regs = cbe_get_cpu_pmd_regs(cpu);

	out_be32 (&pmd_regs->pm_status.val, 0);

	cbe_clear_pm_interrupts(cpu);
}

u32 cbe_query_pm_interrupts(u32 cpu)
{
	struct cbe_pmd_regs __iomem *pmd_regs;
	u32 mask;

	pmd_regs = cbe_get_cpu_pmd_regs(cpu);
	mask = in_be32 (&pmd_regs->pm_status.val);
	return mask;
}

void cbe_enable_pm(u32 cpu)
{
	struct cbe_pmd_regs __iomem *pmd_regs;

	pmd_regs = cbe_get_cpu_pmd_regs(cpu);
	shadow[cpu]->pm_control.ena_perf_mon = 1;
	shadow[cpu]->counter_value_in_latch = 0;
	out_be32 (&pmd_regs->pm_control.val, shadow[cpu]->pm_control.val);
}

void cbe_disable_pm(u32 cpu)
{
	struct cbe_pmd_regs __iomem *pmd_regs;

	pmd_regs = cbe_get_cpu_pmd_regs(cpu);
	shadow[cpu]->pm_control.ena_perf_mon = 0;
	out_be32 (&pmd_regs->pm_control.val, shadow[cpu]->pm_control.val);
}

void cbe_write_pm (u32 cpu, pm_reg_name reg, u32 val)
{
	struct cbe_pmd_regs __iomem *pmd_regs = cbe_get_cpu_pmd_regs(cpu);

	switch (reg) {
		case group_control:
			write_wo_mmio(group_control, val);
			break;

		case debug_bus_control:
			write_wo_mmio(debug_bus_control, val);
			break;

		case pm_control:
			write_wo_mmio(pm_control, val);
			break;

		case pm_interval:
			write_wo_mmio(pm_interval, val);
			break;

		case pm_start_stop:
			write_wo_mmio(pm_start_stop, val);
			break;
	}
}

void cbe_write_pm07_control(u32 cpu, u32 ctr, u32 val)
{
	struct cbe_pmd_regs __iomem *pmd_regs = cbe_get_cpu_pmd_regs(cpu);

	if (ctr < (NR_PHYS_CTRS * 2))
		write_wo_mmio(pm07_control[ctr], val);
}

u32 cbe_read_pm (u32 cpu, pm_reg_name reg)
{
	switch (reg) {
		case group_control:
			return shadow[cpu]->group_control.val;

		case debug_bus_control:
			return shadow[cpu]->debug_bus_control.val;

		case pm_control:
			return shadow[cpu]->pm_control.val;

		case pm_interval:
			return shadow[cpu]->pm_interval.val;

		case pm_start_stop:
			return shadow[cpu]->pm_start_stop.val;

		default:
			return 0;
	}
}

u32 cbe_read_pm07_control(u32 cpu, u32 ctr)
{
	return (ctr < (NR_PHYS_CTRS * 2)) ?
			shadow[cpu]->pm07_control[ctr].val : 0;
}
